package de.ruben.opencms.modules.ftpd;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import de.ruben.opencms.modules.ftpd.Logger;
import de.ruben.opencms.modules.ftpd.ParseException;
import de.ruben.opencms.modules.ftpd.ServerPI;

/**
 * This is the main class for jftpd. It creates the server socket and loops
 * forever accepting user connections and creating ServerPI threads to handle
 * them.
 * 
 * taken from ryan heise's jftpd package: 
 * 	
 * 	http://www.progsoc.uts.edu.au/~rheise/projects/jftpd/
 * 
 * modified by ruben malchow :
 * 
 * 	http://disco.kalkatraz.de/opencmsftpd
 * 
 */

public class Server extends Thread {

	public static final String VERSION = "Version 0.4";
	public static final String BUILD = "Build @@BUILDNUM@@ @@TSTAMP@@";
	private static String chroot;
	private static String suffix;

	public static final int SERVER_PORT = 21;
	public static final int SERVER_DATA_PORT = 20;
	private int port;

	private ServerSocket serverSocket;
	
	private static String[] adminUsers = new String[0];
	
	private boolean running=false;

	public static void main(String[] args) throws Exception {
		if (args.length > 1) {
			System.err.println("Usage: jftpd [ port ]");
			System.exit(1);
		}

		Properties p=new Properties();
		p.put("log.output","stderr");
		p.put("log.priority","");
		Server server = new Server(p, 21);
		server.startServer();
	}

	public Server() throws ParseException {
		this(null,SERVER_PORT);
	}

	public Server(Properties properties) throws ParseException {
		this(properties, SERVER_PORT);
	}

	public Server(Properties properties, int port) throws ParseException {
		if(properties==null) {
			properties=new Properties();
			properties.put("log.output","stderr");
			properties.put("log.priority","");
		}
		configure(properties);
		this.port = port;
	}

	public void run() {
		try {
			serverSocket = new ServerSocket(port);
			System.err.println("OpenCMS FTPD "+ VERSION +" "+BUILD+" waiting for incoming connections on port "+port+" ...");
			while (running) {
				try {
					Socket clientSocket = serverSocket.accept();
					ServerPI pi = new ServerPI(clientSocket);
					Thread t=new Thread(pi);
					t.setPriority(t.getPriority()-1);
					t.start();
					System.err.println("OpenCMS FTPD accepted incoming connection from "+clientSocket.getInetAddress().toString().substring(1));
				} catch (SocketException e) {
					System.err.println("OpenCMS FTPD server socket closed ("+e.getMessage()+")");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		System.err.println("OpenCMS FTPD server loop ended ...");
	}
	
	public void startServer() {
		System.err.println("OpenCMS FTPD starting up ...");
		running=true;
		setDaemon(true);
		start();
	}

	public void stopServer() {
		System.err.println("OpenCMS FTPD shutting down ...");
		running=false;
		try {
			serverSocket.close();
		} catch (IOException se) {
			se.printStackTrace();
		}
		interrupt();
	}

	public void service(Socket socket) throws Exception {
		ServerPI pi = new ServerPI(socket);
		pi.run();
	}

	public static boolean isAdminUser(String user) {
		for (int i = 0; i < adminUsers.length; i++) {
			if(adminUsers[i].compareTo(user)==0) return true;
		}
		return false;
	}
	
	private void configure(Properties properties) throws ParseException {
		
		try {
			
			String priorityStr = properties.getProperty("log.priority");
			
			if (priorityStr == null) {
				throw new ParseException(ParseException.PROPERTY_UNDEFINED, "log.priority", null);
			}
			String priorityNames[] = Logger.priorityNames;
			int priority = Logger.LOG_OFF;
			for (int i = 0; i < priorityNames.length; i++) {
				if (priorityNames[i].compareToIgnoreCase(priorityStr)==0) {
					priority = i;
					break;
				}
			}
	
			Logger.setPriority(priority);
	
			String output = properties.getProperty("log.output");
			OutputStream outputStream = null;
			if (output == null || output.equals("stderr")) {
				outputStream = System.err;
			} else {
				try {
					outputStream = new FileOutputStream(output, true);
				} catch (IOException e) {
					throw new ParseException(ParseException.PROPERTY_INVALID,"log.output", output);
				}
			}
			
			suffix = properties.getProperty("suffix");

			chroot = properties.getProperty("chroot");
			
			String tmp=properties.getProperty("chroot.exceptions");
			if(tmp!=null) {
				StringTokenizer st=new StringTokenizer(tmp,";");
				Vector v=new Vector();
				for(;st.hasMoreElements();) {
					v.addElement(st.nextElement());
				}
				adminUsers = new String[v.size()];
				v.copyInto(adminUsers);
			} else {
				adminUsers = new String[0];
			}
			
			Logger.setWriter(new PrintWriter(new OutputStreamWriter(outputStream), true));
		} catch (Exception e) {
			System.err.println("de.ruben.opencms.modules.ftpd.Server.configure() : ###########################################################");
			System.err.println("de.ruben.opencms.modules.ftpd.Server.configure() : #  ");
			System.err.println("de.ruben.opencms.modules.ftpd.Server.configure() : #  ERROR ");
			System.err.println("de.ruben.opencms.modules.ftpd.Server.configure() : #  ");
			System.err.println("de.ruben.opencms.modules.ftpd.Server.configure() : ###########################################################");
			e.printStackTrace();
		}
	}

	public static String getChroot() {
		return chroot==null?"":chroot;
	}

	public static void setChroot(String chroot) {
		Server.chroot = chroot;
	}

	public static String getSuffix() {
		return suffix==null?"":suffix;
	}

	public static void setSuffix(String suffix) {
		Server.suffix = suffix;
	}
}
