package de.ruben.opencms.modules.ftpd;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.BindException;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketAddress;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.opencms.file.types.CmsResourceTypeBinary;
import org.opencms.file.types.CmsResourceTypeImage;
import org.opencms.file.types.CmsResourceTypeJsp;
import org.opencms.file.types.CmsResourceTypePlain;
import org.opencms.file.types.CmsResourceTypeUnknown;

import de.ruben.opencms.modules.ftpd.CommandException;
import de.ruben.opencms.modules.ftpd.FtpFile;
import de.ruben.opencms.modules.ftpd.Logger;
import de.ruben.opencms.modules.ftpd.Representation;
import de.ruben.opencms.modules.ftpd.ServerPI;
import de.ruben.opencms.modules.ftpd.TransmissionMode;

/**
 * This is the server data transfer process. It is responsible for transferring
 * files to and from the client. A separate data socket is created to transfer
 * the data.
 * 
 * taken from ryan heise's jftpd package: 
 * 	
 * 	http://www.progsoc.uts.edu.au/~rheise/projects/jftpd/
 * 
 * modified by ruben malchow :
 * 
 * 	http://disco.kalkatraz.de/opencmsftpd
 * 
 */

public class ServerDTP {
	
	private static SimpleDateFormat dateFormat = new SimpleDateFormat("MMM dd hh:mm");

	private ServerPI serverPI;
	private String dataHost;
	private int dataPort = -1;
	private TransmissionMode transmissionMode = TransmissionMode.STREAM;
	private Representation representation = Representation.ASCII;
	
	boolean isPassive=false;
	
	private PassiveSocket ps=null;

	public ServerDTP(ServerPI serverPI) {
		this.serverPI = serverPI;
	}

	public void setTransmissionMode(TransmissionMode transmissionMode) {
		this.transmissionMode = transmissionMode;
	}

	public void setDataStructure(char stru) {
		// Ignore. Java itself only supports file-structure, so there
		// is no sense adding record-structure support in the server.
	}

	public Representation getRepresentation() {
		return representation;
	}

	public void setRepresentation(Representation representation) {
		this.representation = representation;
	}

	public void setDataPort(String host, int port) {
		dataHost = host;
		dataPort = port;
	}
	
	private Socket connect() {
		Socket out=null;
		try {
			if ((!isPassive && dataPort == -1)) throw new CommandException(500,"Can't establish data connection: no PORT specified.");
			
			if (!isPassive) {
				out = new Socket(dataHost, dataPort);
				Logger.log(Logger.LOG_DEBUG, "ServerDTP.connect() : ACTIVE DATA CONNECTION "+dataHost+" : "+dataPort);
			} else {
				if(ps==null) throw new CommandException(500,"Can't establish data connection: no PASV socket available.");
				while(out==null) {
					Logger.log(Logger.LOG_DEBUG, "ServerDTP.connect() : waiting for passive connection ...");
					out=ps.getSocket();
					try {
						Thread.sleep(200);
					} catch (Exception e) {
					}
				}
				Logger.log(Logger.LOG_DEBUG, "ServerDTP.connect() : got passive connection!");
			}
			return out;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public int receiveFile(String path) throws CommandException {
		int reply = 0;
		ByteArrayOutputStream fos; 
		Socket dataSocket = null;
		try {
			FtpFile file=new FtpFile(serverPI.getCmsObject(),path);
			if (!file.exists()) {
				Logger.log(Logger.LOG_DEBUG,"ServerDTP.receiveFile() : Creating new file ("+path+")");
				boolean created=file.createNewFile(serverPI.getCmsObject());
				Logger.log(Logger.LOG_DEBUG,"ServerDTP.receiveFile() : Created new file ("+created+")");
				if(!created) {
					throw new CommandException(500,"Can't create file at ("+path+")");
				}
			}

			fos = new ByteArrayOutputStream();

			// Read file contents.
			//
			serverPI.reply(150, "ServerDTP.receiveFile() : Opening " + representation.getName()+ " mode data connection.");
			dataSocket=connect();

			transmissionMode.receiveFile(dataSocket, fos, representation);
			
			byte[] content = fos.toByteArray();
			
			Logger.log(Logger.LOG_DEBUG,"ServerDTP.receiveFile() : received "+content.length+" bytes of data ... ");
			
			synchronized (serverPI) {
				file.setContent(content);
			}
			
			reply = serverPI.reply(226, "Transfer complete.");

		} catch (ConnectException e) {
			e.printStackTrace();
			throw new CommandException(425, "Can't open data connection.");
		} catch (IOException e) {
			e.printStackTrace();
			throw new CommandException(550, "Can't write to file ("+e.getMessage()+")");
		} catch (Exception e) {
			e.printStackTrace();
			if(e instanceof CommandException) throw (CommandException)e;
			throw new CommandException(550, "Can't write to file ("+e.getMessage()+")");
		} finally {
			try {
				if (dataSocket != null) {
					dataSocket.getInputStream().close();
					dataSocket.getOutputStream().flush();
					dataSocket.getOutputStream().close();
					dataSocket.close();
				}				
			} catch (IOException e) {
			}
		}
		return reply;
	}

	public int sendFile(String path) throws CommandException {
		int reply = 0;
		Socket dataSocket = null;
		try {
			FtpFile file=new FtpFile(serverPI.getCmsObject(),path);
			if (!file.exists())
				throw new CommandException(550, "File does not exist in that location.");
			if (!file.isFile()) {
				throw new CommandException(550, "not a plain file");
			}

			dataSocket=connect();
			
			// Send file contents.
			//
			ByteArrayInputStream bais=new ByteArrayInputStream(file.getContent());
			serverPI.reply(150, "Opening " + representation.getName() + " mode data connection.");
			transmissionMode.sendFile(bais, dataSocket, representation);
			reply = serverPI.reply(226, "Transfer complete.");
			
		} catch (FileNotFoundException e) {
			Logger.log(Logger.LOG_DEBUG, "No such file: " + e);
			throw new CommandException(550, "No such file.");
		} catch (ConnectException e) {
			throw new CommandException(425, "Can't open data connection.");
		} catch (IOException e) {
			Logger.log(Logger.LOG_DEBUG, "Not a regular file: " + e);
			throw new CommandException(553, "Not a regular file.");
		} finally {
			try {
				if (dataSocket != null)
					dataSocket.close();
			} catch (IOException e) {
			}
		}
		return reply;
	}

	public int sendNameList(String path) throws CommandException {
		int reply = 0;
		Socket dataSocket = null;
		Logger.log(Logger.LOG_DEBUG, "sendNameList : " + path);
		try {
			FtpFile dir = new FtpFile(serverPI.getCmsObject(),path);
			
			String fileNames[] = dir.list();

			// Connect to User DTP.
			//
			dataSocket = connect();//new Socket(dataHost, dataPort);
			Representation representation = Representation.ASCII;
			PrintWriter writer = new PrintWriter(representation
					.getOutputStream(dataSocket));

			// Send file name list.
			//
			serverPI.reply(150, "Opening " + representation.getName()
					+ " mode data connection.");
			for (int i = 0; i < fileNames.length; i++) {
				writer.print(fileNames[i]);
				writer.print('\n');
			}
			writer.flush();
			reply = serverPI.reply(226, "Transfer complete.");
			Logger.log(Logger.LOG_DEBUG, "end of sendNameList : " + path);
		} catch (ConnectException e) {
			throw new CommandException(425, "Can't open data connection.");
		} catch (Exception e) {
			throw new CommandException(550, "No such directory.");
		} finally {
			try {
				if (dataSocket != null)
					dataSocket.close();
			} catch (IOException e) {
			}
		}
		return reply;
	}

	public int sendList(String path) throws CommandException {
		int reply = 0;
		Logger.log(Logger.LOG_DEBUG, "sendList : " + path);
		Socket dataSocket=null;
		try {


			Representation representation = Representation.ASCII;
			// Send long file list.
			serverPI.reply(150, "Opening " + representation.getName()+" mode data connection.");

			FtpFile dir = new FtpFile(serverPI.getCmsObject(),path);
			Logger.log(Logger.LOG_DEBUG, "ServerDTP.sendList() getting list of children : " + path);
			String fileNames[] = dir.list();
			Logger.log(Logger.LOG_DEBUG, "ServerDTP.sendList() got list : " + fileNames.length+" children");
			int numFiles = fileNames != null ? fileNames.length : 0;

			// connect or wait for connection, whatever mode we're in
			Logger.log(Logger.LOG_DEBUG, "ServerDTP.sendList() : connecting ...");
			dataSocket = connect();
			Logger.log(Logger.LOG_DEBUG, "ServerDTP.sendList() : ... connected");
			PrintWriter writer = new PrintWriter(representation.getOutputStream(dataSocket));

			// Print the total number of files.
			//
			writer.print("total " + numFiles + "\n");

			// Loop through each file and print its name, size,
			// modification date etc.
			//
			for (int i = 0; i < numFiles; i++) {
				
				String fileName = fileNames[i];
				FtpFile file = new FtpFile(serverPI.getCmsObject(), dir.getAbsolutePath()+"/"+fileName);
				listFile(file, writer);
				
				//flush every 20 files to make sure the socket does not time out
				if(i%20==0)writer.flush();
			}

			writer.flush();
			reply = serverPI.reply(226, "Transfer complete.");

			Logger.log(Logger.LOG_DEBUG, "end of sendList : " + path);
		} catch (ConnectException e) {
			throw new CommandException(425, "Can't open data connection.");
		} catch (Exception e) {
			e.printStackTrace();
			throw new CommandException(550, "No such directory.");
		} finally {
			try {
				if (dataSocket != null)
					dataSocket.close();
			} catch (IOException e) {
			}
		}
		return reply;
	}

	/**
	 * Lists a single file in long format (including file sizes and modification
	 * dates etc.).
	 * 
	 * @param file
	 *            the file to list.
	 * @param writer
	 *            the writer to print to.
	 */
	private void listFile(File file, PrintWriter writer) {
		StringBuffer sb=new StringBuffer();

		Date date = new Date(file.lastModified());
		String dateStr = dateFormat.format(date);

		long size = file.length();
		String sizeStr = Long.toString(size);
		int sizePadLength = Math.max(8 - sizeStr.length(), 0);
		String sizeField = pad(sizePadLength) + sizeStr;

		sb.append(file.isDirectory() ? 'd' : '-');
		
		sb.append("rwxrwxrwx");
		sb.append(" ");
		sb.append("  1");
		sb.append(" ");
		sb.append("ftp     ");
		sb.append(" ");
		sb.append("ftp     ");
		sb.append(" ");
		sb.append(sizeField);
		sb.append(" ");
		sb.append(dateStr);
		sb.append(" ");
		sb.append(file.getName());

		sb.append('\n');
		
		writer.print(sb.toString());
		
		
	}

	private static String pad(int length) {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < length; i++)
			buf.append((char) ' ');
		return buf.toString();
	}

	public boolean isPassive() {
		return isPassive;
	}

	public void setPort() {
		isPassive=false;
	}
	
	public int setPassive(InetAddress bind) {
		PassiveSocket ps=new PassiveSocket(bind);
		isPassive=true;
		this.ps=ps;
		return ps.getPort();
	}
	
	public class PassiveSocket {
		
		private Socket socket;
		private ServerSocket sSocket;
		private SocketAddress sSocketAddress;
		private InetAddress bind;
		private int port=-1;
		
		private boolean bound=false;
		
		public PassiveSocket(InetAddress _bind) {
			
			this.bind=_bind;

			sSocket=null;
			sSocketAddress=null;
			while (!bound) {
				int tmpPort=0;
				try {
					tmpPort=49152+(int)Math.floor(Math.random()*15848d);
					sSocketAddress=new InetSocketAddress(bind,tmpPort);
					sSocket=new ServerSocket();
					Logger.log(Logger.LOG_DEBUG, "PassiveSocket.<init> : PASSIVE DATA CONNECTION trying "+bind+":"+tmpPort);
					sSocket.bind(sSocketAddress,1);
					port=tmpPort;
					bound=true;
				} catch (BindException be) {
					Logger.log(Logger.LOG_DEBUG, "PassiveSocket.<init> : PASSIVE DATA CONNECTION port "+tmpPort+" is in use ("+be.getMessage()+")");
				} catch (Exception e) {
					Logger.log(Logger.LOG_DEBUG, "PassiveSocket.<init> : PASSIVE DATA CONNECTION an exception occured ("+e.getMessage()+")");
					break;
				}
			}
			
			if(bound) {
	
				Logger.log(Logger.LOG_DEBUG, "PassiveSocket.<init> : PASSIVE DATA CONNECTION bound on "+sSocketAddress.toString());
	
				Runnable r=new Runnable() {
					public void run() {
						
						try {
							//InetAddress remoteAddress=InetAddress.getByName(dataHost);
							Logger.log(Logger.LOG_DEBUG, "PassiveSocket.$1.run() : PASSIVE DATA CONNECTION accepting ...");
							socket = sSocket.accept();
							Logger.log(Logger.LOG_DEBUG, "PassiveSocket.$1.run() : PASSIVE DATA CONNECTION ... accepted!");
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				};
				Thread t=new Thread(r);
				t.setDaemon(true);
				t.start();
			} else {
				Logger.log(Logger.LOG_DEBUG, "PassiveSocket.<init> : PASSIVE DATA CONNECTION could not bind, aborting ");
			}
		}

		public int getPort() {
			return port;
		}

		public void setPort(int port) {
			this.port = port;
		}

		public Socket getSocket() {
			return socket;
		}

	}
	
	
}
