package de.ruben.opencms.modules.ftpd;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.Socket;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.opencms.file.CmsObject;
import org.opencms.file.types.CmsResourceTypeFolder;
import org.opencms.main.OpenCms;

import de.ruben.opencms.modules.ftpd.CommandException;
import de.ruben.opencms.modules.ftpd.FtpFile;
import de.ruben.opencms.modules.ftpd.Logger;
import de.ruben.opencms.modules.ftpd.Representation;
import de.ruben.opencms.modules.ftpd.Server;
import de.ruben.opencms.modules.ftpd.ServerDTP;
import de.ruben.opencms.modules.ftpd.TransmissionMode;

/**
 * This is the server protocol interpreter. The main Server object creates one
 * of these for each user connection to interpret user commands.
 * 
 * This class implements the Runnable interface so that it can be run as a
 * thread or invoked directly.
 * 
 * taken from ryan heise's jftpd package: 
 * 	
 * 	http://www.progsoc.uts.edu.au/~rheise/projects/jftpd/
 * 
 * modified by ruben malchow :
 * 
 * 	http://disco.kalkatraz.de/opencmsftpd
 * 
 */

public class ServerPI implements Runnable {
	
	private CmsObject co;
	
	private ServerDTP dtp;

	private Socket clientSocket;
	private BufferedReader reader;
	private PrintWriter writer;
	
	private Class commandHandlerArgTypes[] = { String.class, StringTokenizer.class };

	private String username;
	private String password;
	private String baseDir = Server.getChroot();
	private String currentDir = baseDir;
	
	private boolean isAdminUser=false;

	public ServerPI(Socket clientSocket) throws IOException {
		this.clientSocket = clientSocket;
		reader = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
		writer = new PrintWriter(new OutputStreamWriter(new BufferedOutputStream(clientSocket.getOutputStream())), true);
		dtp = new ServerDTP(this);
	}

	public void run() {
		try {
			clientLoop();
		} catch (Exception e) {
			e.printStackTrace();
			Logger.log(Logger.LOG_EMERG, "clientLoop failed: " + e);
		} finally {
			try {
				clientSocket.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	private void clientLoop() throws Exception {
		
		//reply(214, "welcome .... ");
		//reply(220, "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
		reply(220, clientSocket.getLocalAddress().toString().substring(1)+" (OpenCMS FTPD" + Server.VERSION + ")");
		//reply(220, "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
		//reply(220, "localhost(FTP server ready)");

		String line = null;
		while ((line = reader.readLine()) != null) {

			StringTokenizer st = new StringTokenizer(line);
			String command = st.nextToken().toLowerCase();

			if ("PASS".equalsIgnoreCase(command)) {
				Logger.log(Logger.LOG_DEBUG, ">>> PASS ********");
			} else {
				Logger.log(Logger.LOG_DEBUG, ">>> "+line);
			}
			Object args[] = { line, st };
			try {
				int code = -1;
				if(command.compareTo("xpwd")==0) command="pwd";
				Method commandHandler = getClass().getMethod("handle_" + command, commandHandlerArgTypes);
				code = ((Integer) commandHandler.invoke(this, args)).intValue();
				if (code == 221) return;
			} catch (OutOfMemoryError oops) {
				oops.printStackTrace();
			} catch (InvocationTargetException e) {
				try {
					Throwable t=e.getTargetException();
					if(t instanceof Exception) {
						throw (Exception)e.getTargetException();
					} else {
						t.printStackTrace();
						throw new RuntimeException(t);
					}
				} catch (CommandException ce) {
					reply(ce.getCode(), ce.getText());
				} catch (NoSuchElementException e1) {
					reply(500, "'" + line + "': command not understood.");
				}
			} catch (NoSuchMethodException e) {
					reply(500, "'" + line + "': command not understood.");
			} catch (Exception e) {
				Logger.log(Logger.LOG_ERR, "Exception invoking " + command + " command handler: " + e);
				e.printStackTrace();
			} catch (Throwable t) {
				throw new Exception(t.getMessage());
			}
		}
	}

	public int handle_user(String line, StringTokenizer st) throws CommandException {
		username = st.nextToken();
		return reply(331, "Password required for " + username + ".");
	}

	public int handle_pass(String line, StringTokenizer st) throws CommandException {
		if (username == null)
			throw new CommandException(503, "Login with USER first.");

		String password = null;
		if (st.hasMoreTokens()) {
			password = st.nextToken();
		} else {
			password = "";
		}

		try {
			co=OpenCms.initCmsObject("Guest");
			co.loginUser(this.username,password);
			this.password = password;
			isAdminUser = Server.isAdminUser(this.username);
			return reply(230, "User " + username + " logged in.");
		} catch (Exception e) {
			try {
				Thread.sleep(Math.round(Math.random()*3000));
			} catch (Exception e2) {
			}
			return reply(530, "Bad Username/Password");
		}
	}

	private int handle_unsupported(String line) throws CommandException {
		checkLogin();
		throw new CommandException(500, "'" + line+ "': command not supported.");
	}
	
	public int handle_acct(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}
	
	public int handle_cwd(String line, StringTokenizer st) throws CommandException {

		checkLogin();

		String newDir = st.nextToken();
		newDir = resolvePath(newDir);
		//newDir = normalizePath(resolvePath(newDir));

		FtpFile file=new FtpFile(co,newDir);
		
		if (!file.exists()){
			throw new CommandException(550, newDir + ": no such directory");
		}
		if (!file.isDirectory()){
			throw new CommandException(550, newDir + ": not a directory");
		}
		currentDir = newDir;
		Logger.log(Logger.LOG_DEBUG, "new cwd = " + currentDir);
		return reply(250, "CWD command successful.");
	}

	public int handle_cdup(String line, StringTokenizer st) throws CommandException {
		String s=resolvePath(currentDir+"/");
		s=s.substring(0,s.lastIndexOf("/"));
		currentDir=resolvePath(s);
		return reply(250, "CDUP command successful.");
	}

	public int handle_smnt(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_quit(String line, StringTokenizer st) throws CommandException {
		username = null;
		password = null;
		return reply(221, "Goodbye.");
	}

	public int handle_rein(String line, StringTokenizer st) throws CommandException {

		checkLogin();

		username = null;
		password = null;
		currentDir = "/";
		dtp = new ServerDTP(this);
		co=null;
		return reply(220, "Service ready for new user.");
	}

	public int handle_port(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String portStr = st.nextToken();
		st = new StringTokenizer(portStr, ",");
		String h1 = st.nextToken();
		String h2 = st.nextToken();
		String h3 = st.nextToken();
		String h4 = st.nextToken();
		int p1 = Integer.parseInt(st.nextToken());
		int p2 = Integer.parseInt(st.nextToken());

		String dataHost = h1 + "." + h2 + "." + h3 + "." + h4;
		int dataPort = (p1 << 8) | p2;

		dtp.setDataPort(dataHost, dataPort);
		dtp.setPort();

		return reply(200, "PORT command successful.");
	}

	public int handle_pasv(String line, StringTokenizer st) throws CommandException {
		checkLogin();

		dtp.setDataPort(clientSocket.getInetAddress().toString().substring(1),-1);
		
		int port=dtp.setPassive(clientSocket.getLocalAddress());
		
		String portStr = clientSocket.getLocalAddress().toString();
		st = new StringTokenizer(portStr, ".");
		int p1=(int)Math.floor(port/256);
		int p2=port%256;


		String h=clientSocket.getLocalAddress().toString();
		
		h=h.replace('.',',');
		h=h.replace('/',' ');
		h=h.trim();
		
		String dataHost = h + ","+p1+","+p2;
		return reply(227, "Passive command successful (" + dataHost+")");
	}

	public int handle_type(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String arg = st.nextToken().toUpperCase();
		if (arg.length() != 1) throw new CommandException(500, "TYPE: invalid argument '" + arg + "'");
		char code = arg.charAt(0);
		Representation representation = Representation.get(code);
		if (representation == null) throw new CommandException(500, "TYPE: invalid argument '" + arg+ "'");
		dtp.setRepresentation(representation);
		return reply(200, "Type set to " + arg);
	}

	public int handle_stru(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String arg = st.nextToken().toUpperCase();
		try {
			if (arg.length() != 1) throw new Exception();
			char stru = arg.charAt(0);
			switch (stru) {
			case 'F':
				dtp.setDataStructure(stru);
				break;
			default:
				throw new Exception();
			}
		} catch (Exception e) {
			throw new CommandException(500, "STRU: invalid argument '" + arg + "'");
		}
		return reply(200, "STRU " + arg + " ok.");
	}

	public int handle_mode(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String arg = st.nextToken().toUpperCase();
		if (arg.length() != 1)throw new CommandException(500, "MODE: invalid argument '" + arg+ "'");
		char code = arg.charAt(0);
		TransmissionMode mode = TransmissionMode.get(code);
		if (mode == null)throw new CommandException(500, "MODE: invalid argument '" + arg+ "'");
		dtp.setTransmissionMode(mode);
		return reply(200, "MODE " + arg + " ok.");
	}

	public int handle_retr(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String path = null;
		try {
			path = line.substring(5);
		} catch (Exception e) {
			throw new NoSuchElementException(e.getMessage());
		}

		path = resolvePath(path);
		Logger.log(Logger.LOG_DEBUG, "handle_retr: " + path);

		return dtp.sendFile(path);
	}

	public int handle_stor(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String path = null;
		try {
			path = line.substring(5);
			path = path.replace(' ','_');
			path = path.replace(',','.');
		} catch (Exception e) {
			throw new NoSuchElementException(e.getMessage());
		}
		path = resolvePath(path);
		return dtp.receiveFile(path);
	}

	public int handle_stou(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_appe(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_allo(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_rest(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_rnfr(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_rnto(String line, StringTokenizer st)throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_abor(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_dele(String line, StringTokenizer st) throws CommandException {
		checkLogin();

		String arg = st.nextToken();

		String filePath = resolvePath(arg);

		FtpFile file=new FtpFile(co,filePath);
		
		if(!file.exists()) {
			throw new CommandException(550, arg + ": file does not exist");
		} else {
			try {
				if(!file.delete()) {
					throw new CommandException(550, arg + ": could not delete file");
				}
			} catch (Exception e) {
				throw new CommandException(550, arg + ": could not delete file ("+e.getMessage()+")");
			}
		}
		return reply(250, "DELE command successful.");
	}

	public int handle_rmd(String line, StringTokenizer st) throws CommandException {

		checkLogin();

		String arg = st.nextToken();

		String filePath = resolvePath(arg);

		FtpFile file=new FtpFile(co,filePath);
		
		if(!file.exists()) {
			throw new CommandException(550, arg + ": directory does not exist");
		} else {
			try {
				file.delete();
			} catch (Exception e) {
				throw new CommandException(550, arg + ": could not delete directory ("+e.getMessage()+")");
			}
		}

		
		return reply(250, "RMD command successful.");
	}

	public int handle_mkd(String line, StringTokenizer st) throws CommandException {
		checkLogin();

		String arg = st.nextToken();

		String dirPath = resolvePath(arg);

		FtpFile file=new FtpFile(co,dirPath);
		
		if(file.exists()) {
			throw new CommandException(550, arg + ": file exists");
		} else {
			String path=dirPath+"/";
			path=ServerPI.normalizePath(path);
			String dir=path.substring(0,path.lastIndexOf("/"));
			String name=path.substring(path.lastIndexOf("/")+1);
			path=ServerPI.normalizePath(path);
			try {
				FtpFile ff=new FtpFile(co,dir);
				ff.createChild(name,CmsResourceTypeFolder.getStaticTypeId());
			} catch (Exception e) {
				e.printStackTrace();
				throw new CommandException(550, arg + ": directory could not be created ("+e.getMessage()+")");
			}
		}
		return reply(257, "\"" + dirPath + "\" directory created");
	}

	public int handle_pwd(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		currentDir=resolvePath(currentDir);
		return reply(257, "\""+currentDir+"\"");
	}

	public int handle_list(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String path = null;
		while(st.hasMoreTokens()) {
			String s=st.nextToken();
			if(!s.startsWith("-")) {
				path=s;
				break;
			}
		}
		
		if(path==null) {
			path=currentDir;
		}

		path = resolvePath(path);

		return dtp.sendList(path);
	}

	public int handle_nlst(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		String path = null;
		if (st.hasMoreTokens()) 
			path = st.nextToken();
		else
			path = currentDir;

		path = resolvePath(path);

		return dtp.sendNameList(path);
	}

	public int handle_site(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_syst(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_stat(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_help(String line, StringTokenizer st) throws CommandException {
		return handle_unsupported(line);
	}

	public int handle_noop(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		return reply(200, "NOOP command successful.");
	}

	// 
	// UNDOCUMENTED COMMANDS
	//


	public int handle_size(String line, StringTokenizer st) throws CommandException {

		checkLogin();

		String arg = st.nextToken();
		String path = resolvePath(arg);
		
		FtpFile file = new FtpFile(co,path);
		if(!file.exists()) {
			throw new CommandException(550, arg + ": no such file");
		} else if(!file.isFile()) {
			throw new CommandException(550, arg + ": not a plain file");
		} else {
		}

		long size;
		try {
			Representation representation = dtp.getRepresentation();
			size = representation.sizeOf(file);
		} catch (IOException e) {
			throw new CommandException(550, e.getMessage());
		}

		// XXX: in ascii mode, we must count newlines properly
		return reply(213, "" + size);
	}

	public int handle_mdtm(String line, StringTokenizer st) throws CommandException {
		checkLogin();

		String arg = st.nextToken();
		String path = resolvePath(arg);
		FtpFile file = new FtpFile(co,path);

		if (!file.exists())
			throw new CommandException(550, arg + ": no such file");

		Date date = new Date(file.lastModified());
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddhhmmss");
		String dateStr = dateFormat.format(date);

		return reply(213, dateStr);
	}

	
	/*
	 * 
	 * CUSTOM COMMANDS
	 * 
	 */

	public int handle_prop(String line, StringTokenizer st) throws CommandException {
		checkLogin();

		return 0;
	}

	public int handle_types() throws CommandException {
		checkLogin();


		return 0;
	}

	public int handle_create(String line, StringTokenizer st) throws CommandException {
		checkLogin();
		return 0;
	}

	/**
	 * Writes a reply line to the client socket.
	 * 
	 * @param code
	 *            the reply code.
	 * @param text
	 *            the string message to include in the reply.
	 * @return the reply code for this reply.
	 */
	int reply(int code, String text) {
		try {
			writer.print(code + " " + text+"\r\n");
			Logger.log(Logger.LOG_DEBUG, "<<< "+code+" "+text);
			writer.flush();
			return code;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return -1;
	}

	

	/**
	 * Resolves an FTP given by the client. Relative paths will be resolved
	 * relative to currentDir. '.' path segments will be removed, and '..' path
	 * segments will pop the previous segment of the path stack (if there is a
	 * previous segment).
	 */
	String resolvePath(String path) throws CommandException {

		if(path.length()==0) path="/";

		if (path.charAt(0) != '/') {
			path = '/' + currentDir + "/" + path;
		} else {
			path = '/' + path;
		}
		
		StringTokenizer pathSt = new StringTokenizer(path, "/");
		Stack segments = new Stack();
		while (pathSt.hasMoreTokens()) {
			String segment = pathSt.nextToken();
			if (segment.equals("..")) {
				if (!segments.empty())
					segments.pop();
			} else if (segment.equals(".")) {
				// skip
			} else {
				segments.push(segment);
			}
		}

		StringBuffer pathBuf = new StringBuffer("/");
		Enumeration segmentsEn = segments.elements();
		while (segmentsEn.hasMoreElements()) {
			pathBuf.append(segmentsEn.nextElement());
			if (segmentsEn.hasMoreElements())
				pathBuf.append("/");
		}

		path = normalizePath(pathBuf.toString());
		
		if(isAdminUser) {
		} else if(!path.startsWith(baseDir)) {
			path=baseDir;
			throw new CommandException(550,"you can't go there");
		}

		return path;
	}

	public static String normalizePath(String path) {
		Pattern p=Pattern.compile("//");
		Matcher m=p.matcher(path);
		path=m.replaceAll("/");
		if(path.endsWith("/") && path.length()>1) {
			path=path.substring(0,path.length()-1);
		}
		return path;
	}
	
	void checkLogin() throws CommandException {
		if (password == null) throw new CommandException(530, "Please login with USER and PASS.");
	}
	
	public static File createCmsFile(CmsObject co, String path) {
		try {
			File f=new FtpFile(co,path);
			return f;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	public CmsObject getCmsObject() {
		return co;
	}

}
